import java.awt.image.BufferedImage;
import java.util.List;
import java.util.ArrayList;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.Rectangle;
import java.awt.image.ImageObserver;

/*
 * Test images for calibration:
 * - "IGVCclip1 003.jpg" (long distance, excellent light)
 * 
 * - "IGVCclip2 131.jpg" (medium distance, poor light, tiny acute solar hilight)
 *   "IGVCclip2 140.jpg" (ditto)
 * 
 * - "IGVCclip3 001.jpg" (medium distance, excellent light)
 * 
 * - "IGVCclip4 033.jpg" (short distance, good light, large strong solar hilight)
 * 
 * - "IGVCclip4 173.jpg" (short distance, good light)
 * 
 * - "IGVCclip4 297.jpg" (short distance, good light, large moderate solar shadow)
 * 
 * - "IGVCclip5 066.jpg" (short distance, good light, large strong solar hilight)
 * 
 * - "IGVCclip5 066.jpg" (short distance, poor light, lens flare)
 */

/**
 * @author David Foster <davidfstr@gmail.com>
 */
// EXPERIMENTAL
public class BarrelBlobFinder implements PixelBufferToBufferedImageFilter, ParameterizedFilter {
	private final static Color BARREL_BOUND_COLOR = Color.YELLOW;
	private final static int BARREL_BOUND_THICKNESS = 3;
	private final static Color BARREL_COUNTER_COLOR = Color.YELLOW;
	
	private final static boolean DEBUG_UPPER_EDGE_DETERMINATION = false;
	
	/* lower values tolerate less orange-pixel noise but give better results
	 * when little orange-pixel noise is present */
	// XXX: chosen default value is arbitrary; subsequent adjustment recommended
	private final FloatFilterParam barrelEdgeNoiseToleranceParam =
		new FloatFilterParam("Barrel Edge - Noise Tolerance", 0.00f, 0.50f, 0.10f);
	
	/* higher values look for brighter orange */
	// XXX: chosen default value is arbitrary; subsequent adjustment recommended
	private final IntFilterParam orangePixelDetectionThresholdParam =
		new IntFilterParam("Orange Pixels (Red-Green) - Threshold (High)", -255, 255, 30);
	
	/* lower values look for cleaner (less dirty) white */
	// >=90 for low-light conditions
	private final IntFilterParam whitePixelSaturationThresholdParam =
		new IntFilterParam("White Pixel Saturation - Threshold (Low)", 0, 255, 80);			
	
	/* higher values look for brighter white */
	// >=40 to eliminate black
	private final IntFilterParam whitePixelBrightnessThresholdParam =
		new IntFilterParam("White Pixel Brightness - Threshold (High)", 0, 255, 40);
	
	/* higher values allow longer gaps between orange/white stripes,
	 * tolerate larger highlights on the barrel,
	 * tolerate larger deep shadows on the barrel,
	 * but distinguish individual barrels in barricades less accurately
	 * 
	 * NOTE: Be careful not to set this parameter too high (especially when calibrating other parameters).
	 */
	// <=15 to avoid overriding white-stripe detection
	// >=2 for low-light conditions (with shadows)
	private final IntFilterParam maxStripeGapLengthParam =
		new IntFilterParam("Maximum Stripe Gap Length", 0, Integer.MAX_VALUE, 5);
	
	/* lower values look for thinner barrels, but are more confused by orange-pixel noise */
	// <=60 to detect far away barrels
	private final IntFilterParam numOrangePixelsInBarrelGirthThresholdParam =
		new IntFilterParam("Minimum Barrel Girth (Width + Noise)", 1, Integer.MAX_VALUE, 40);
	
	/* lower values allow detection of very near and very far barrels,
	 * but make it easier to incorrectly identify small *sections* of
	 * barrels as being full-sized barrels */
	// <=90 to detect far away barrels
	private final IntFilterParam minBarrelHeightParam =
		new IntFilterParam("Minimum Barrel Height", 1, Integer.MAX_VALUE, 90);
	
	/* higher values detect the upper-edge of barrels more accurately in the presense
	 * of multiple barrels, but tolerate less tapering */
	// XXX: chosen default value is arbitrary; subsequent adjustment recommended
	private final FloatFilterParam taperToleranceParam =
		new FloatFilterParam("Barrel Tapering % - Tolerated Threshold (High)", 0.00f, 1.00f, 0.70f);
	
	/* higher values detect the width of barrels more accurately */
	// XXX: chosen default value is arbitrary; subsequent adjustment recommended
	private final IntFilterParam maxHeightOfBarrelCurvatureParam =
		new IntFilterParam("Maximum Height of Barrel Curvature", 1, Integer.MAX_VALUE, 10);
	
	private final FilterParam[] params = new FilterParam[] {
		barrelEdgeNoiseToleranceParam,
		orangePixelDetectionThresholdParam,
		whitePixelSaturationThresholdParam,
		whitePixelBrightnessThresholdParam,
		maxStripeGapLengthParam,
		numOrangePixelsInBarrelGirthThresholdParam,
		minBarrelHeightParam,
		taperToleranceParam,
		maxHeightOfBarrelCurvatureParam
	};
	
	// ### FILTER ###
	
	private int[] pixelOrangeness = null;
	private int[] pixelSaturation = null;
	private int[] pixelBrightness = null;
	
	private final List barrelBoundList = new ArrayList();	// List<Rectangle>
	
	public void filter(int width, int height, int[] inPixels,
	                   BufferedImage inImg, BufferedImage outImg)
	{
		/* Extract values of filter parameters */
		float barrelEdgeNoiseTolerance = barrelEdgeNoiseToleranceParam.getValue();
		int orangePixelDetectionThreshold = orangePixelDetectionThresholdParam.getValue();
		int maxStripeGapLength = maxStripeGapLengthParam.getValue();
		int numOrangePixelsInBarrelGirthThreshold = numOrangePixelsInBarrelGirthThresholdParam.getValue();
		int minBarrelHeight = minBarrelHeightParam.getValue();
		float taperTolerance = taperToleranceParam.getValue();
		int maxHeightOfBarrelCurvature = maxHeightOfBarrelCurvatureParam.getValue();
		
		int numPixels = inPixels.length;
		
		/*
		 * Allocate/resize intermediate filter buffers as needed
		 */
		if ((this.pixelOrangeness == null) || (this.pixelOrangeness.length != numPixels)) {
			this.pixelOrangeness = new int[numPixels];
			this.pixelSaturation = new int[numPixels];
			this.pixelBrightness = new int[numPixels];
		}
		
		/* 
		 * Calculate the "orangeness", saturation, and brightness of each pixel:
		 * > pixelOrangeness[x,y] = (int) (getRed(pixel[x,y]) - getGreen(pixel[x,y]))
		 * > pixelSaturation[x,y] = (int) getSaturation(ARGBToAHSB(pixel[x,y]))
		 * > pixelBrightness[x,y] = (int) getBrightness(ARGBToAHSB(pixel[x,y]))
		 */
		for (int i=0; i<numPixels; i++) {
			int curPixelARGB = inPixels[i];
			pixelOrangeness[i] = Pixels.getRed(curPixelARGB) - Pixels.getGreen(curPixelARGB);
			
			int curPixelAHSB = Pixels.ARGBToAHSB(curPixelARGB);
			pixelSaturation[i] = Pixels.getSaturation(curPixelAHSB);
			pixelBrightness[i] = Pixels.getBrightness(curPixelAHSB);
		}
		
		/* Begin scanning rows of the image for lower-edges of barrels, starting from the bottom */
		int numBarrelsDetected;
		{
			for (int y=height-1, rowOff=numPixels-width;
				 y>=0;
				 y--, rowOff-=width)
			{
				/* Count the number of "orange" pixels in the current row */
				int numOrangePixelsInRow = countOrangePixels(rowOff, width);
				
				/* 
				 * If the number of orange pixels in the current row
				 * exceeds the barrel-detection threshold, assume that we have
				 * encountered the lower-edge of one or more barrels, and
				 * attempt to identify those barrels.
				 */
				if (numOrangePixelsInRow > numOrangePixelsInBarrelGirthThreshold) {
					int lowerBarrelEdgeY = y;
					
					/* 
					 * XXX: The following lower-barrel-edge isolation algorithm is simplified, 
					 *      in that it makes the assumption that the lower edges of more than
					 *      one barrel will never occur in the same row
					 * 
					 * Locate the lower edge of the barrel, "shaving" off any orange-pixel "noise".
					 * 
					 * Locate the (barrelEdgeNoiseTolerance)% and the
					 * (1.0-barrelEdgeNoiseTolerance)% percentiles of the list of
					 * orange pixels in the current row and the "nearby" rows above it.
					 * (A row is considered "nearby" if it is within (maxHeightOfBarrelCurvature)
					 *  rows of the current row.)
					 * 
					 * REMARK: The reason that "nearby" rows are assessed in addition to the
					 *         current row is that this yields a significantly more accurate
					 *         value for the width of the detected barrel in instances when
					 *         the bottom of the barrel appears curved (which is very common).
					 * 
					 * The interval of pixels between the *average* values of these
					 * percentiles will be designated as comprising the lower-edge
					 * of the barrel.
					 */
					int avgLowPercentileOrangePixelX;
					int avgHighPercentileOrangePixelX;
					{
						int sumLowPercentileOrangePixelX = 0;
						int sumHighPercentileOrangePixelX = 0;
						
						int numNearbyRows = Math.min(maxHeightOfBarrelCurvature, y+1);
						int limitY2 = y - numNearbyRows;		// NOTE (when porting to C/C++): min value is -1; this is a signed variable
						for (int y2 = y, rowOff2 = rowOff;
						     y2 > limitY2;
						     y2--, rowOff2-=width)
						{
							int numOrangePixelsInRow2 = countOrangePixels(rowOff2, width);
							
							int lowPercentileOrangePixelX = 0;			// (default)
							int highPercentileOrangePixelX = width-1;	// (default)
							{
								int lowPercentileOrangePixelOrdinal = 
									(int) ((numOrangePixelsInRow2-1) * barrelEdgeNoiseTolerance);
								int highPercentileOrangePixelOrdinal =
									(int) ((numOrangePixelsInRow2-1) - lowPercentileOrangePixelOrdinal);
								for (int x=0, off=rowOff, orangePixCount=0; x<width; x++, off++) {
									if (pixelOrangeness[off] > orangePixelDetectionThreshold) {
										orangePixCount++;
										
										if (orangePixCount == lowPercentileOrangePixelOrdinal) {
											lowPercentileOrangePixelX = x;
										}
										if (orangePixCount == highPercentileOrangePixelOrdinal) {
											highPercentileOrangePixelX = x;
										}
									}
								}
							}
							
							sumLowPercentileOrangePixelX += lowPercentileOrangePixelX;
							sumHighPercentileOrangePixelX += highPercentileOrangePixelX;
						}
						
						avgLowPercentileOrangePixelX = sumLowPercentileOrangePixelX/numNearbyRows;
						avgHighPercentileOrangePixelX = sumHighPercentileOrangePixelX/numNearbyRows;
					}
					int leftBarrelEdgeX = avgLowPercentileOrangePixelX;
					int rightBarrelEdgeX = avgHighPercentileOrangePixelX;
					int barrelWidth = rightBarrelEdgeX - leftBarrelEdgeX + 1;
					
					/*
					 * Locate the upper edge of the barrel
					 * (while tolerating a certain amount of tapering and
					 *  while tolerating gaps up to a certain length).
					 * 
					 * Scan row-segments above the detected lower-barrel-edge,
					 * starting from the bottom and moving upward.
					 * 
					 * Stop scanning when either:
					 * 1) more than (maxStripeGapLength) row-segments are
					 *    encountered whose pixels are less than 
					 *    (taperTolerance)% orange/white 
					 * 2) the top of the image is encountered
					 */
					int upperBarrelEdgeY;
					{
						int lastSeenStripeUpperEdgeY = lowerBarrelEdgeY;		// (default)
						int numOrangeOrWhitePixelsInRowSegmentThreshold
							= (int) (barrelWidth * taperTolerance);
						for (int y2=lowerBarrelEdgeY, rowSegOff = rowOff+leftBarrelEdgeX, gapLength = 0;
							 (y2>=0) && (gapLength <= maxStripeGapLength);
							 y2--, rowSegOff-=width)
						{
							int numOrangePixelsInRowSegment = countOrangePixels(rowSegOff, barrelWidth);
							int numWhitePixelsInRowSegment = countWhitePixels(rowSegOff, barrelWidth);
							int numOrangeOrWhitePixelsInRowSegment = 
								numOrangePixelsInRowSegment +
								numWhitePixelsInRowSegment;
							
							if (numOrangeOrWhitePixelsInRowSegment < numOrangeOrWhitePixelsInRowSegmentThreshold) {
								/*
								 * Insufficent number of orange/white pixels found;
								 * we've either:
								 *  1) hit the upper-edge of the barrel
								 *  2) hit a gap between adjacent stripes
								 */
								gapLength++;
								
								if (DEBUG_UPPER_EDGE_DETERMINATION) {
									System.out.println(
										"DEBUG: "+numOrangeOrWhitePixelsInRowSegment+
											"("+numOrangePixelsInRowSegment+","+
											numWhitePixelsInRowSegment+") < "+
										numOrangeOrWhitePixelsInRowSegmentThreshold+" < "+
										barrelWidth);
								}
							} else {
								// Reset gap length to zero (since we no longer see a gap)
								gapLength = 0;
								
								// Record the position of the stripe that we currently see
								lastSeenStripeUpperEdgeY = y2;
							}
						}
						upperBarrelEdgeY = lastSeenStripeUpperEdgeY;
					}
					int barrelHeight = lowerBarrelEdgeY - upperBarrelEdgeY + 1;
					
					/*
					 * Ensure the identified "barrel" satisfies the minimum height requirement
					 */
					if (barrelHeight >= minBarrelHeight) {
						/*
						 * In subsequent scans, ignore the region of the image in which
						 * we identified the barrel.
						 */
						for (int y2=upperBarrelEdgeY, rowOff2 = (upperBarrelEdgeY*width) + leftBarrelEdgeX;
							 y2<=lowerBarrelEdgeY;
							 y2++, rowOff2+=width)
						{
							for (int x2=leftBarrelEdgeX, off2=rowOff2;
								 x2<=rightBarrelEdgeX;
								 x2++, off2++)
							{
								pixelOrangeness[off2] = -255;
								pixelSaturation[off2] = 255;
								pixelBrightness[off2] = 0;
							}
						}
						
						// Record the bounds of the barrel that was found
						Rectangle barrelBounds = new Rectangle(
							leftBarrelEdgeX,	// x
							upperBarrelEdgeY,	// y
							barrelWidth,		// width
							barrelHeight		// height
						);
						this.barrelBoundList.add(barrelBounds);
					}
				}
			}
			
			numBarrelsDetected = this.barrelBoundList.size();
		}
		
		/* 
		 * Prepare output image.
		 * 
		 * The output image consists of the input image,
		 * plus annotations that show where the boundary of each detected barrel is,
		 * plus a counter in the lower-left corner that displays the number of barrels detected.
		 */
		{
			Graphics g = outImg.getGraphics();
			g.drawImage(inImg, 0, 0, (ImageObserver) null);
			
			g.setColor(BARREL_BOUND_COLOR);
			for (int i=0, n=this.barrelBoundList.size(); i<n; i++) {
				Rectangle curBarrelBound = (Rectangle) this.barrelBoundList.get(i);
				
				int barrelX = curBarrelBound.x;
				int barrelY = curBarrelBound.y;
				int barrelWidth = curBarrelBound.width;
				int barrelHeight = curBarrelBound.height;
				for (int j=BARREL_BOUND_THICKNESS; j>=0; j--) {
					/* 
					 * NOTE: The -1's on the width and height are to compensate for the way 
					 * that drawRect(...) determines the positions of the right/bottom
					 * edges of the rectangle to draw
					 */
					g.drawRect(barrelX, barrelY, barrelWidth-1, barrelHeight-1);

					barrelX++; barrelY++; barrelWidth-=2; barrelHeight-=2;
				}
			}
			this.barrelBoundList.clear();		// recycle list
			
			g.setColor(BARREL_COUNTER_COLOR);
			g.drawString(String.valueOf(numBarrelsDetected), 0, height-1);
		}
		
		if (DEBUG_UPPER_EDGE_DETERMINATION) {
			System.out.println("DEBUG: done");
		}
	}
	
	private int countOrangePixels(int off, int len) {
		int orangePixelDetectionThreshold = orangePixelDetectionThresholdParam.getValue();
		
		int numOrangePixelsInRow = 0;
		for (; len>0; len--, off++) {
			if (pixelOrangeness[off] >= orangePixelDetectionThreshold) {
				numOrangePixelsInRow++;
			}
		}
		return numOrangePixelsInRow;
	}
	
	private int countWhitePixels(int off, int len) {
		int whitePixelSaturationThreshold = whitePixelSaturationThresholdParam.getValue();
		int whitePixelBrightnessThreshold = whitePixelBrightnessThresholdParam.getValue();
		
		int numWhitePixelsInRow = 0;
		for (; len>0; len--, off++) {
			if ((pixelSaturation[off] <= whitePixelSaturationThreshold) &&
				(pixelBrightness[off] >= whitePixelBrightnessThreshold))
			{
				numWhitePixelsInRow++;
			}
		}
		return numWhitePixelsInRow;
	}
	
	public String toString() {
		return "Barrel Blob Finder";
	}
	
	// ### PARAMETERIZED FILTER ###
	
	public FilterParam[] getParams() {
		return params;
	}
}