public class HorizontalContourExtractionFilter implements PixelBufferFilter2, ParameterizedFilter {
	private final IntFilterParam stippleRadiusParam
		= new IntFilterParam("Stipple Radius", 1, Integer.MAX_VALUE, 5);
	private final BooleanFilterParam directionalParam
		= new BooleanFilterParam("Directional", false);
	
	private final FilterParam[] params = new FilterParam[] {
		stippleRadiusParam,
		directionalParam
	};
	
	// ### INIT ###
	
	public HorizontalContourExtractionFilter() {
		this(5, false);
	}
	
	/**
	 * @param stippleRadius		specifies the radius of the smallest blob 
	 * 							whose contours the filter can find effectively;
	 * 							(use smaller values to detect the contours of thinner objects)
	 * @param directional		specifies whether this filter distinguishes between
	 * 							"orientation" of a contour (light-to-dark vs. dark-to-light)
	 */
	public HorizontalContourExtractionFilter(int stippleRadius, boolean directional) {
		this.stippleRadiusParam.setValue(stippleRadius);
		this.directionalParam.setValue(directional);
	}
	
	// ### FILTER ###
	
	public void filter(int width, int height, int[] inPixels, int[] outPixels) {
		int stippleRadius = this.stippleRadiusParam.getValue();
		boolean directional = this.directionalParam.getValue();
		
		int off = 0;
		int edgeLeftLimit = stippleRadius;
		int edgeRight = width-stippleRadius;
		for (int y=0; y<height; y++) {
			for (int x=0; x<edgeLeftLimit; x++, off++) {
				outPixels[off] = directional
					? Pixels.packPixel(255,128,128,128)
					: Pixels.packPixel(255,0,0,0);
			}
			
			for (int x=edgeLeftLimit; x<edgeRight; x++, off++) {
				// Take sum of the channels on the left side of the stipple's center pixel
				int sumOfRedOnFirstSide = 0;
				int sumOfGreenOnFirstSide = 0;
				int sumOfBlueOnFirstSide = 0;
				for (int i=stippleRadius, off2=off-1; i>0; i--, off2--) {
					int curPixel = inPixels[off2];
					sumOfRedOnFirstSide += Pixels.getRed(curPixel);
					sumOfGreenOnFirstSide += Pixels.getGreen(curPixel);
					sumOfBlueOnFirstSide += Pixels.getBlue(curPixel);
				}
				
				// Take sum of the channels on the right side of the stipple's center pixel
				int sumOfRedOnSecondSide = 0;
				int sumOfGreenOnSecondSide = 0;
				int sumOfBlueOnSecondSide = 0;
				for (int i=stippleRadius, off2=off+1; i>0; i--, off2++) {
					int curPixel = inPixels[off2];
					sumOfRedOnSecondSide += Pixels.getRed(curPixel);
					sumOfGreenOnSecondSide += Pixels.getGreen(curPixel);
					sumOfBlueOnSecondSide += Pixels.getBlue(curPixel);
				}
				
				// Calculate the average (per pixel) difference between the channels 
				// for the pixels on either side of the stipple's center pixel
				int avgDeltaOfRed = (sumOfRedOnSecondSide - sumOfRedOnFirstSide)/stippleRadius;
				int avgDeltaOfGreen = (sumOfGreenOnSecondSide - sumOfGreenOnFirstSide)/stippleRadius;
				int avgDeltaOfBlue = (sumOfBlueOnSecondSide - sumOfBlueOnFirstSide)/stippleRadius;
				
				// Normalize deltas, which are in the range [-255, 255], to be in the range [0,255]
				// (The method of normalization depends on whether this filter is directional or not.)
				int normalizedDeltaOfRed, normalizedDeltaOfGreen, normalizedDeltaOfBlue;
				if (directional) {
					normalizedDeltaOfRed = 128 + avgDeltaOfRed/2;
					normalizedDeltaOfGreen = 128 + avgDeltaOfGreen/2;
					normalizedDeltaOfBlue = 128 + avgDeltaOfBlue/2;
				} else {
					normalizedDeltaOfRed = Math.abs(avgDeltaOfRed);
					normalizedDeltaOfGreen = Math.abs(avgDeltaOfGreen);
					normalizedDeltaOfBlue = Math.abs(avgDeltaOfBlue);
				}
				
				outPixels[off] = Pixels.packPixel(
					255,							// alpha
					normalizedDeltaOfRed,			// red
					normalizedDeltaOfGreen,			// green
					normalizedDeltaOfBlue			// blue
				);
			}
			
			for (int x=edgeRight; x<width; x++, off++) {
				outPixels[off] = directional
					? Pixels.packPixel(255,128,128,128)
					: Pixels.packPixel(255,0,0,0);
			}
		}
	}
	
	public String toString() {
		return "Contour Extraction (Horizontal)";
	}
	
	// ### PARAMETERIZED FILTER ###
	
	public FilterParam[] getParams() {
		return params;
	}
}